/* Copyright (C) 2003 Datapark corp. All rights reserved.
   Copyright (C) 2000-2002 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "dps_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <dirent.h>

#include "dps_common.h"
#include "dps_hash.h"
#include "dps_guesser.h"
#include "dps_utils.h"
#include "dps_unicode.h"
#include "dps_log.h"
#include "dps_vars.h"
#include "dps_mutex.h"

/* This should be last include */
#ifdef DMALLOC
#include "dmalloc.h"
#endif

/*#define DEBUG_GUESSER*/

static DPS_LANGMAP *FindLangMap(DPS_LANGMAPLIST *L, char *lang, char *charset, char *filename) {
  DPS_LANGMAP *o = NULL;
  register size_t i;
  char *lt, *language = dps_strtok_r(lang, ", ", &lt);

  while(language) {
    for(i = 0; i < L->nmaps; i++) {
      if(
	 (!strcasecmp(L->Map[i].charset, charset))
	 &&
	 (!strcasecmp(L->Map[i].lang, language))
	 ) {
	return &L->Map[i];
      }
    }
    language = dps_strtok_r(NULL, ", ", &lt);
  }

  if(L->nmaps == 0){
    o = L->Map = (DPS_LANGMAP*)DpsMalloc(sizeof(DPS_LANGMAP));
    if (o == NULL) return NULL;
  }else{
    L->Map = (DPS_LANGMAP*)DpsRealloc(L->Map, (L->nmaps + 1) * sizeof(DPS_LANGMAP));
    if (L->Map == NULL) {
      L->nmaps = 0;
      return NULL;
    }
    o = &L->Map[L->nmaps];
  }
  if (o == NULL || L->Map == NULL) {
    fprintf(stderr, "Can't alloc/realloc for language map (%s, %s), nmaps: %d (%d)", lang, charset, 
		 L->nmaps + 1, (L->nmaps + 1) * sizeof(DPS_LANGMAP) );
    return NULL;
  }
  bzero((void*)o, sizeof(DPS_LANGMAP));
  for (i = 0; i <= DPS_LM_HASHMASK; i++) o->memb[i].index = i;
  o->charset = (char*)DpsStrdup(charset);
  o->lang = (char*)DpsStrdup(lang);
  o->filename = (filename == NULL) ? NULL : (char*)DpsStrdup(filename);
  L->nmaps++;
  return o;
}

__C_LINK int __DPSCALL DpsLoadLangMapFile(DPS_LANGMAPLIST *L, const char * filename) {
     FILE * f;
     char str[1000];
     char *Ccharset = NULL, *Clanguage = NULL;
     DPS_LANGMAP *Cmap = NULL;
     DPS_CHARSET * cs;

     f=fopen(filename,"r");
     if(!f){
          fprintf(stderr, "Can't open LangMapFile '%s'\n", filename);
          return DPS_ERROR;
     }
     while(fgets(str,sizeof(str),f)){
          if(str[0]=='#'||str[0]==' '||str[0]=='\t')continue;

          if(!strncmp(str,"Charset:",8)){
               
               char * charset, * lasttok;
               DPS_FREE(Ccharset);
               if((charset = dps_strtok_r(str+8," \t\n\r",&lasttok))){
                       const char *canon = DpsCharsetCanonicalName(charset);
                    if (canon) {
                      Ccharset = (char*)DpsStrdup(canon);
                    } else {
                      fprintf(stderr, "Charset: %s in %s not supported\n", charset, filename);
                      return DPS_ERROR;
                    }
               }
          }else
          if(!strncmp(str,"Language:",9)){
               char * lang, *lasttok;
               DPS_FREE(Clanguage);
               if((lang = dps_strtok_r(str+9," \t\n\r",&lasttok))){
                    Clanguage = (char*)DpsStrdup(lang);
               }
          }else{
               char *s;
               int count;
               
               if(!(s=strchr(str,'\t')))continue;
               if(Clanguage == NULL) {
                 fprintf(stderr, "No language definition in LangMapFile '%s'\n", filename);
                 return DPS_ERROR;
               }

               if(Ccharset == NULL) {
                 fprintf(stderr, "No charset definition in LangMapFile '%s'\n", filename);
                 return DPS_ERROR;
               }
               if(!(cs = DpsGetCharSet(Ccharset))) {
                 fprintf(stderr, "Unknown charset '%s' in LangMapFile '%s'\n", Ccharset, filename);
                 return DPS_ERROR;
               }
               if (Cmap == NULL) {
		 Cmap = FindLangMap(L, Clanguage, Ccharset, filename);
		 DpsSort(Cmap->memb, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
	       }
               if (Cmap == NULL) return DPS_ERROR;
               *s='\0';

               if(((count=atoi(s+1))==0)||(strlen(str)<1)||(strlen(str)>DPS_LM_MAXGRAM))
                    continue;

               for(s=str;*s;s++){
                    if(*s=='_')*s=' ';
               }
               if(*str){
                    register int hindex;
                    hindex = ((unsigned int)DpsStrHash32(str)) & DPS_LM_HASHMASK;
                    Cmap->memb[hindex].count += count;
                    strcpy(Cmap->memb[hindex].str, str);
               }
          }
     }
     fclose(f);
     DPS_FREE(Clanguage);
     DPS_FREE(Ccharset);
     
     if (Cmap != NULL) DpsPrepareLangMap(Cmap);
     return DPS_OK;
}

__C_LINK int __DPSCALL DpsLoadLangMapList(DPS_LANGMAPLIST *L, const char * mapdir) {

	DIR * dir;
	struct dirent * item;
	char fullname[1024]="";
	char name[1024]="";
	int res = DPS_OK;

	dir=opendir(mapdir);
	if(!dir)return 0;

	while((item=readdir(dir))){
		char * tail;
		strcpy(name,item->d_name);
		if((tail=strstr(name,".lm"))){
		        *tail='\0';
			sprintf(fullname,"%s/%s",mapdir,item->d_name);
			res = DpsLoadLangMapFile(L, fullname);
			if (res != DPS_OK) return res;
		}
	}
	closedir(dir);

	return DPS_OK;
}


void DpsLangMapListFree(DPS_LANGMAPLIST *List){
     size_t i;
     
     for(i=0;i<List->nmaps;i++){
          DPS_FREE(List->Map[i].charset);
          DPS_FREE(List->Map[i].lang);
          DPS_FREE(List->Map[i].filename);
     }
     DPS_FREE(List->Map);
     List->nmaps = 0;
}

void DpsLangMapListSave(DPS_LANGMAPLIST *List) {
     size_t i, j, minv;
     FILE *out;
     DPS_LANGMAP *Cmap;
     char name[128];
     
     for(i = 0; i < List->nmaps; i++) {
       Cmap = &List->Map[i];
       if (Cmap->needsave) {
         if (Cmap->filename == NULL) {
           dps_snprintf(name, 128, "%s.%s.lm", Cmap->lang, Cmap->charset);
           if ((out = fopen(name, "w")) == NULL) continue;
         } else {
           if ((out = fopen(Cmap->filename, "w")) == NULL) continue;
         }
         fprintf(out, "#\n");
         fprintf(out, "# Autoupdated.\n");
         fprintf(out, "#\n\n");
         fprintf(out, "Language: %s\n", Cmap->lang);
         fprintf(out, "Charset:  %s\n", Cmap->charset);
         fprintf(out, "\n\n");
         DpsSort(Cmap->memb, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpCount);
	 minv = (Cmap->memb[DPS_LM_TOPCNT - 1].count > 1000) ? 1000 : Cmap->memb[DPS_LM_TOPCNT - 1].count;
         for(j = 0; j < DPS_LM_TOPCNT; j++) {
           Cmap->memb[j].count -= Cmap->memb[DPS_LM_TOPCNT - 1].count - minv;
         }
         for(j = 0; j < DPS_LM_TOPCNT; j++) {
          char * s;
          if(!Cmap->memb[j].count) break;
          
          for(s = Cmap->memb[j].str; *s; s++)
               if(*s == ' ') *s='_';
          
          fprintf(out, "%s\t%d\n", Cmap->memb[j].str, Cmap->memb[j].count);
         }
         fclose(out);
       }
     }
}

typedef struct {
  unsigned char buf[2 * DPS_LM_MAXGRAM + 1];
  size_t p, s;
} DPS_PAT;

void DpsBuildLangMap(DPS_LANGMAP * map,const char * text,size_t textlen, int StrFlag) {
  DPS_PAT Pat[DPS_LM_MAXGRAM];
  const char * end = text + textlen;
  unsigned char code, prevb = 32;
  register size_t i;
  unsigned int hindex;

  for (i = 0; i < DPS_LM_MAXGRAM; i++) {
    memset(Pat[i].buf, 32, 2 * DPS_LM_MAXGRAM + 1);
    Pat[i].buf[2 * DPS_LM_MAXGRAM] = 0;
    Pat[i].p = 0;
    Pat[i].s = i;
  }
  for (; text <= end; text++) {
    code = (unsigned char)(*text);
    if (code < 32) continue;
    if ((code == 32) && (prevb == 32)) continue;
/*    if (code < ' ') continue;
    if ((code == ' ') && (prevb == ' ')) continue;*/
    prevb = code;
  
    for (i = 0; i < DPS_LM_MAXGRAM; i++) {
      Pat[i].buf[Pat[i].s] = Pat[i].buf[Pat[i].s + i + 1] = code;
      hindex = DpsHash32(Pat[i].buf + Pat[i].p, i + 1);
      hindex = ((unsigned int)hindex ) & DPS_LM_HASHMASK;
      map->memb[hindex].count++;
#ifdef DEBUG_GUESSER
      /* Print collision */
      if(map->memb[hindex].str[0]) {
	int res;
	res = strncmp(map->memb[hindex].str, Pat[i].buf + Pat[i].p, i + 1);
	if(res){
	  printf("Coll %04X '%s' '%s'\n",hindex,map->memb[hindex].str, Pat[i].buf + Pat[i].p);
	  strncpy(map->memb[hindex].str, Pat[i].buf + Pat[i].p, i + 1);
	}
      }
#endif
      if (StrFlag) {
	strncpy(map->memb[hindex].str, Pat[i].buf + Pat[i].p, i + 1);
	map->memb[hindex].str[i + 1] = '\0';
      }
/*      if (i == 1) fprintf(stderr, "i:%d  p:%d  s:%d [%s]\n", i, Pat[i].p, Pat[i].s, Pat[i].buf);*/
      Pat[i].p++; if (Pat[i].p > i) Pat[i].p = 0;
      Pat[i].s++; if (Pat[i].s > i) Pat[i].s = 0;
    }
  }
}



int DpsLMstatcmp(const void * i1, const void * i2){
  register const DPS_MAPSTAT *s1 = (const DPS_MAPSTAT *)i1;
  register const DPS_MAPSTAT *s2 = (const DPS_MAPSTAT *)i2;
  if (s1->hits > s2->hits) return -1;
  if (s1->hits < s2->hits) return 1;
  if (s1->miss < s2->miss) return -1;
  if (s1->miss > s2->miss) return 1;

  return 0;
}

int DpsLMcmpCount(const void * i1,const void * i2){
     const DPS_LANGITEM * m1=i1;
     const DPS_LANGITEM * m2=i2;
     
     if (m2->count > m1->count) return 1;
     if (m2->count < m1->count) return -1;
     return 0;
}

int DpsLMcmpIndex(const void * i1,const void * i2) {
     const DPS_LANGITEM * m1=i1;
     const DPS_LANGITEM * m2=i2;
     
     if (m2->index > m1->index) return -1;
     if (m2->index < m1->index) return 1;
     return 0;
}


void DpsPrepareLangMap(DPS_LANGMAP * map){

     DpsSort(map->memb, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpCount);
     DpsSort(map->memb, DPS_LM_TOPCNT, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
}


void DpsCheckLangMap(DPS_LANGMAP * map0, DPS_LANGMAP * map1, DPS_MAPSTAT *stat, size_t InfMiss) {
     register int i;
     DPS_LANGITEM *HIT;

     stat->hits = stat->miss = 0;
     for (i = 0; i < DPS_LM_TOPCNT; i++) {
       if ( (HIT = bsearch(&map1->memb[i], map0->memb, DPS_LM_TOPCNT, sizeof(DPS_LANGITEM), &DpsLMcmpIndex)) == NULL) {
	 stat->miss++;
       } else {
	 stat->hits += DPS_LM_TOPCNT - abs(HIT - map0->memb);
       }
       if (stat->miss > InfMiss) break;
     }
}


int  DpsGuessCharSet(DPS_AGENT *Indexer, DPS_DOCUMENT * Doc,DPS_LANGMAPLIST *List,DPS_LANGMAP *LangMap){
     DPS_CHARSET * cs;
     DPS_MAPSTAT * mapstat = NULL;
     const char *lang = DpsVarListFindStr(&Doc->Sections, "Content-Language", "");
     const char *meta_lang = DpsVarListFindStr(&Doc->Sections, "Meta-Language", "");
     const char *server_charset = DpsVarListFindStr(&Doc->Sections, "Server-Charset", "");
     const char *meta_charset = DpsVarListFindStr(&Doc->Sections, "Meta-Charset", "");
     const char *charset = DpsVarListFindStr(&Doc->Sections, "RemoteCharset", "");
     size_t InfMiss = DPS_LM_TOPCNT + 1;
     size_t i;
     int have_server_lang = (*lang != '\0');
     int use_meta, update_lm;
     
     use_meta = Indexer->Flags.use_meta;
     update_lm = Indexer->Flags.update_lm;
			
     if (*charset == '\0') {
       if (*server_charset != '\0' 
	   && (!use_meta || *meta_charset == '\0' || !strcasecmp(server_charset, meta_charset) )) charset = server_charset;
       else if (use_meta && *server_charset == '\0' && *meta_charset != '\0') charset = meta_charset;
     }
     DpsVarListReplaceStr(&Doc->Sections, "Charset", charset);

     if (*lang == '\0') {
       if (*meta_lang != '\0' && use_meta) DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = meta_lang);
     }
     
     /* 
          TODO for Guesser
          
          There are three sources to detect charset:
          1. HTTP header:  Content-Type: ... charset=XXX
          2. <META HTTP-EQUIV="Content-Type" Contnet="... charset=YYY">
          3. ZZZ[i] - array of guessed charsets in mapstat[]
          good(ZZZ[n]) means that guesser returned good results for n.

          Now we may have various combinations:
          Simple situations, non-guessed and guessed charsets
          seem to be the same. At least one of non-guessed
          charset is the same with the best guessed charset
          and guessed charset gave good results:

          1. XXX==YYY==ZZZ[0],      good(ZZZ[0]). Take XXX value.
          2. No YYY, XXX=ZZZ[0] and good(ZZZ[0]). Take XXX value.
          3. No XXX, YYY=ZZZ[0] and good(ZZZ[0]). Take YYY value.
          4. XXX<>YYY, XXX==ZZZ[0], good(ZZZ[0]). Take XXX value.
          5. XXX<>YYY, YYY==ZZZ[0], good(ZZZ[0]). Take XXX value.
               4 and 5 seem to be webmaster mistake.

          There are the same fith situations when ZZZ[x] is still good 
          enough, but it is not the best one, i.e. x>0 
          Choose charset in the same way.
     */

     if (*charset == '\0' || *lang == '\0') {

       /* Prepare document langmap */
       DpsPrepareLangMap(LangMap);

       /* Allocate memory for comparison statistics */
       if ((mapstat=(DPS_MAPSTAT *)DpsMalloc((List->nmaps + 1) * sizeof(DPS_MAPSTAT))) == NULL) {
	 DpsLog(Indexer, DPS_LOG_ERROR, "Can't alloc momory for DpsGuessCharSet (%d bytes)", List->nmaps*sizeof(DPS_MAPSTAT));
	 return DPS_ERROR;
       }
     
       DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
       for(i=0;i<List->nmaps;i++){
          mapstat[i].map=&List->Map[i];
          if ((*charset == '\0') && (*lang == '\0')) {
            DpsCheckLangMap(&List->Map[i], LangMap, &mapstat[i], InfMiss);
          } else if ((*charset) && 
                  (!strcasecmp(mapstat[i].map->charset, charset) || !strcasecmp(mapstat[i].map->charset, meta_charset))  ) {
            DpsCheckLangMap(&List->Map[i], LangMap, &mapstat[i], InfMiss);
          } else if ((*lang) && !strncasecmp(mapstat[i].map->lang, lang, strlen(mapstat[i].map->lang) )) {
            DpsCheckLangMap(&List->Map[i], LangMap, &mapstat[i], InfMiss);
          } else {
            mapstat[i].hits = 0;
	    mapstat[i].miss = DPS_LM_TOPCNT + 1;
          }
          if (mapstat[i].miss < InfMiss) InfMiss = mapstat[i].miss;
       }
       DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);

       /* Sort statistics in quality order */
       if (List->nmaps > 1) DpsSort(mapstat, List->nmaps, sizeof(DPS_MAPSTAT), &DpsLMstatcmp);

       /* Display results, best is shown first */
/*
       for (i = 0; i < 5; i++)
	 fprintf(stderr, "Guesser: %dh:%dm %s-%s\n",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);
*/

       if (*server_charset != '\0' || *meta_charset != '\0')
	 for(i=0;i<List->nmaps;i++){

	   if (mapstat[i].map->lang && (*lang != '\0') && (!strncasecmp(mapstat[i].map->lang, lang, strlen(mapstat[i].map->lang) ))) {
	     if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, server_charset)) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
	     } else if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, meta_charset)) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
	     }
	   } else {
	     if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, server_charset)) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	     } else if(mapstat[i].map->charset && !strcasecmp(mapstat[i].map->charset, meta_charset)) {
	       DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
	     }
	   }
	   if (*charset != '\0') break;
	 }
          
       for(i=0;i<List->nmaps;i++){

          if ((*lang != '\0') && (*charset != '\0')) break;
          if(mapstat[i].map->lang && *lang == '\0'/* && (*charset == '\0' || (!strcasecmp(mapstat[i].map->charset, charset)))*/ ){
               DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[i].map->lang);
          }
          if (mapstat[i].map->charset && *charset == '\0' && (!strcmp(lang, mapstat[i].map->lang)) ) {
               DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[i].map->charset);
          }
          
/*          fprintf(stderr, "Guesser: %dh:%dm %s-%s\n",mapstat[i].hits, mapstat[i].miss,mapstat[i].map->lang,mapstat[i].map->charset);*/
          
       }
       if (List->nmaps > 0 && mapstat[0].map->charset && (*charset == '\0') ) {
	     DpsVarListReplaceStr(&Doc->Sections, "Charset", charset = mapstat[0].map->charset);
       }
       if (List->nmaps > 0 && mapstat[0].map->lang && (*lang == '\0') ) {
             DpsVarListReplaceStr(&Doc->Sections, "Content-Language", lang = mapstat[0].map->lang);
       }
     }

     if (have_server_lang && (*server_charset != '\0' || (*meta_charset != '\0' && use_meta ))) {
       if (update_lm) {
         DPS_LANGMAP *Cmap;

	 DPS_GETLOCK(Indexer,DPS_LOCK_CONF);
         if ((Cmap = FindLangMap(&Indexer->Conf->LangMaps, lang, charset, NULL)) != NULL) {
	   DpsSort(Cmap->memb, DPS_LM_HASHMASK + 1, sizeof(DPS_LANGITEM), &DpsLMcmpIndex);
           for (i = 0; i < DPS_LM_HASHMASK+1; i++) {
	     if (Cmap->memb[i].count == 0) {
	       strcpy(Cmap->memb[i].str, LangMap->memb[i].str);
	     }
	     Cmap->memb[i].count += LangMap->memb[i].count;
           }
           DpsPrepareLangMap(Cmap);
           Cmap->needsave = 1;
	   DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
           DpsLog(Indexer, DPS_LOG_EXTRA, "Lang map: %s.%s updated", Cmap->lang, Cmap->charset);
         } else {
	   DPS_RELEASELOCK(Indexer,DPS_LOCK_CONF);
	 }
       }
     }
     cs = DpsGetCharSet(DpsVarListFindStr(&Doc->Sections, "Charset", "iso8859-1"));
     if (cs) Doc->charset_id = cs->id;
     DPS_FREE(mapstat);
     return DPS_OK;
}
